import { Subscription, Observable, ReplaySubject, debounce, from, asapScheduler, Subscriber, fromEvent, distinctUntilChanged } from 'rxjs';
import { Deferred, isSetEqual, isTruthy } from '@s-libs/js-core';
import { flow, bindKey, keys, map as map$1 } from '@s-libs/micro-dash';
import { switchMap, finalize, delay, filter, tap, map, withLatestFrom, scan, startWith } from 'rxjs/operators';

/**
 * Log the values emitted from any observable to the [redux devtools extension](https://github.com/reduxjs/redux-devtools).
 *
 * @param options These are passed along to the extension as is. See its documentation [here](https://github.com/reduxjs/redux-devtools/blob/main/extension/docs/API/Arguments.md).
 */
function logToReduxDevtoolsExtension(observable, options) {
    const extension = window
        .__REDUX_DEVTOOLS_EXTENSION__;
    if (!extension) {
        return new Subscription();
    }
    const connection = extension.connect(options);
    return observable.subscribe((value) => {
        connection.send({}, value);
    });
}

/**
 * 1. Caches the last value emitted to give to new subscribers (without running any upstream pipe operators)
 * 2. Manages all subscribers directly, without passing subscriptions up the stream.
 *
 * This is very similar to `shareReplay(1)`, except that once all subscribers unsubscribe this also unsubscribes from the upstream observable.
 *
 * ```ts
 * const source = new BehaviorSubject(1000);
 * const result = source.pipe(expensiveComputation(), cache());
 * source.subscribe(); // expensiveComputation(1000) runs
 * source.subscribe(); // the cached result is used
 * source.next(2000); // expensiveComputation(2000) runs once, emitted to both subscribers
 * ```
 */
function cache() {
    return (source) => {
        let middleMan;
        let upstreamSubscription;
        return new Observable((subscriber) => {
            if (!middleMan) {
                middleMan = new ReplaySubject(1);
                upstreamSubscription = source.subscribe(middleMan);
            }
            const subscription = middleMan.subscribe(subscriber);
            // teardown logic
            return () => {
                subscription.unsubscribe();
                if (middleMan.observers.length === 0) {
                    upstreamSubscription.unsubscribe();
                    middleMan = undefined;
                }
            };
        });
    };
}

/**
 * It's like {@linkcode https://rxjs-dev.firebaseapp.com/api/operators/exhaustMap exhaustMap}, except it debounces upstream emissions until the previous result completes.
 *
 * ```
 * source:                     -0-1-2-------3-|
 *                              i--B-|
 *                                   i--B-|
 *                                          i--B-|
 * debounceMap((i) => i--B-|): -0--B-2--B---3--B-|
 */
function debounceMap(map) {
    let lastOperationComplete = Promise.resolve();
    return flow(debounce(async () => lastOperationComplete), switchMap((value) => {
        const deferred = new Deferred();
        lastOperationComplete = deferred.promise;
        return from(map(value)).pipe(finalize(bindKey(deferred, 'resolve')));
    }));
}

/**
 * Delays the emission of items from the source Observable using the microtask queue.
 */
function delayOnMicrotaskQueue() {
    return delay(0, asapScheduler);
}

/**
 * Allows items through whose keys are distinct from the previous item.
 *
 * ```
 * source:                     |-{a:1,b:2}--{a:2,b:3}--{a:2,c:3}--{b:3}--{b:4}-|
 * distinctUntilKeysChanged(): |-{a:1,b:2}-------------{a:2,c:3}--{b:3}--------|
 * ```
 */
function distinctUntilKeysChanged() {
    let lastKeySet;
    return filter((value) => {
        const keySet = new Set(keys(value));
        if (lastKeySet && isSetEqual(keySet, lastKeySet)) {
            return false;
        }
        lastKeySet = keySet;
        return true;
    });
}

/**
 * Use this to create a complex pipeable operator. It is usually better style to compose existing operators than to create a brand new one, but when you need full control this can reduce some boilerplate.
 *
 * The supplied `subscriber` will act as a simple pass-through of all values, errors, and completion to `destination`. Modify it for your needs.
 *
 * A simple example, recreating the "map" operator:
 * ```ts
 * function map<I, O>(fn: (input: I) => O) {
 *   return createOperatorFunction<I, O>(
 *     (subscriber, destination) => {
 *       subscriber.next = (value) => {
 *         destination.next(fn(value));
 *       };
 *     },
 *   );
 * }
 * ```
 *
 * For a more complex example, check the source of `skipAfter`.
 */
function createOperatorFunction(modifySubscriber) {
    return (source) => new Observable((destination) => {
        const subscriber = new Subscriber(destination);
        modifySubscriber(subscriber, destination);
        return source.subscribe(subscriber);
    });
}

/**
 * Works like `filter()`, but always lets through the first emission for each new subscriber. This makes it suitable for subscribers that expect the observable to behave like a `BehaviorSubject`, where the first emission is processed synchronously during the call to `subscribe()`.
 *
 * ```
 * source:                   |-false--true--false--true--false--true-|
 * filterBehavior(identity): |-false--true---------true---------true-|
 * filterBehavior(identity):        |-true---------true---------true-|
 * filterBehavior(identity):              |-false--true---------true-|
 * ```
 */
function filterBehavior(predicate) {
    return createOperatorFunction((subscriber, destination) => {
        let firstValue = true;
        subscriber.next = (value) => {
            if (firstValue) {
                destination.next(value);
                firstValue = false;
                return;
            }
            try {
                if (predicate(value)) {
                    destination.next(value);
                }
            }
            catch (ex) {
                destination.error(ex);
            }
        };
    });
}

/**
 * Logs values, errors and completion to the console, and passes them all along unchanged.
 *
 * ```ts
 * of(1, 2).pipe(logValues()).subscribe();
 * // prints using console.log:
 * // [value] 1
 * // [value] 2
 * // [complete]
 *
 * of(1, 2).pipe(logValues("my number", "debug")).subscribe();
 * // prints using console.debug:
 * // [value] my number 1
 * // [value] my number 2
 * // [complete] my number
 *
 * throwError("boo").pipe(logValues("pipe says", "warn")).subscribe();
 * // prints using console.warn:
 * // [error] pipe says boo
 * ```
 */
function logValues(prefix, level = 'log') {
    return tap({
        next: makeLogFn('[value]'),
        error: makeLogFn('[error]'),
        complete: makeLogFn('[complete]'),
    });
    function makeLogFn(...prefixes) {
        if (prefix !== undefined) {
            prefixes.push(prefix);
        }
        return (...values) => {
            console[level](...prefixes, ...values);
        };
    }
}

function mapAndCacheElements(buildCacheKey, buildDownstreamItem) {
    let cache = new Map();
    return map((upstreamItems) => {
        const nextCache = new Map();
        const downstreamItems = map$1(upstreamItems, (upstreamItem, key) => {
            const cacheKey = buildCacheKey(upstreamItem, key);
            let downstreamItem;
            if (cache.has(cacheKey)) {
                downstreamItem = cache.get(cacheKey);
            }
            else if (nextCache.has(cacheKey)) {
                downstreamItem = nextCache.get(cacheKey);
            }
            else {
                downstreamItem = buildDownstreamItem(upstreamItem, key);
            }
            nextCache.set(cacheKey, downstreamItem);
            return downstreamItem;
        });
        cache = nextCache;
        return downstreamItems;
    });
}

/**
 * Applies `buildDownstreamItem` to each item in the upstream array and emits the result. Each downstream item is cached using the key generated by `buildCacheKey` so that the next emission contains references to the matching objects from the previous emission, without running `buildDownstreamItem` again. The cache is only held between successive emissions.
 *
 * This is useful e.g. when using the result in an `*ngFor` expression of an angular template, to prevent angular from rebuilding the inner component and to allow `OnPush` optimizations in the inner component.
 *
 * If multiple items in an upstream array have the same cache key, it will only call `buildDownstreamItem` once.
 *
 * ```ts
 * const mapWithCaching = mapAndCacheArrayElements(
 *   (item) => item,
 *   (item) => item + 1
 * )
 * ```
 * ```
 * source:         -[1, 2]---[1, 2, 3]---[2]--|
 * mapWithCaching: -[2, 3]---[2, 3, 4]---[3]--|
 * ```
 */
const mapAndCacheArrayElements = mapAndCacheElements;

/**
 * Applies `buildDownstreamItem` to each item in the upstream object and emits an array containing the results. Each downstream item is cached using the key generated by `buildCacheKey` so that the next emission contains references to the matching objects from the previous emission, without running `buildDownstreamItem` again. The cache is only held between successive emissions.
 *
 * This is useful e.g. when using the result in an `*ngFor` expression of an angular template, to prevent angular from rebuilding the inner component and to allow `OnPush` optimizations in the inner component.
 *
 * If multiple items in an upstream object have the same cache key, it will only call `buildDownstreamItem` once.
 *
 * ```ts
 * const mapWithCaching = mapAndCacheObjectElements(
 *   (item, key) => key,
 *   (item, key) => item + 1
 * )
 * ```
 * ```
 * source:         -{ a: 1, b: 2 }---{ a: 1, b: 2, c: 3 }---{ b: 2 }--|
 * mapWithCaching: -[2, 3]-----------[2, 3, 4]--------------[3]--|
 * ```
 */
const mapAndCacheObjectElements = mapAndCacheElements;

/**
 * Emits the latest value of the given Observable every time the source Observable emits a value.
 *
 * ```
 * source:               -1---2--3------4-|
 * inner:                ---a------b--c---|
 * mapToLastFrom(inner): -----a--a------c-|
 * ```
 */
function mapToLatestFrom(inner$) {
    return flow(withLatestFrom(inner$), map(([_, inner]) => inner));
}

/**
 * Causes the next value in the pipe to be skipped after `skip$` emits a value. For example:
 *
 * ```
 * source: -1-----2-----3-----4-----5-|
 * skip$:  ----0----------0-0----------
 *
 * result: -1-----------3-----------5-|
 * ```
 * ```ts
 * const source = new Subject();
 * const skip$ = new Subject();
 * const result = source.pipe(skipAfter(skip$));
 *
 * source.next(1); // result emits `1`
 *
 * skip$.next();
 * source.next(2); // result does not emit
 * source.next(3); // result emits `3`
 *
 * skip$.next();
 * skip$.next();
 * source.next(4); // result does not emit
 * source.next(5); // result emits `5`
 * ```
 */
function skipAfter(skip$) {
    return createOperatorFunction((subscriber, destination) => {
        let skipNext = false;
        subscriber.add(skip$.subscribe({
            next: () => {
                skipNext = true;
            },
            error: bindKey(destination, 'error'),
        }));
        subscriber.next = (value) => {
            if (skipNext) {
                skipNext = false;
            }
            else {
                destination.next(value);
            }
        };
    });
}

function withHistory(count) {
    return scan((buf, value) => [value, ...buf.slice(0, count)], []);
}

/**
 * Creates an observable that emits when the [page visibility]{@link https://developer.mozilla.org/en-US/docs/Web/API/Page_Visibility_API} changes. It also emits the current visibility immediately upon subscription.
 *
 * ```ts
 * isPageVisible$().subscribe((isVisible) => {
 *   if (isVisible) {
 *     console.log('Page is visible');
 *   } else {
 *     console.log('Page is hidden');
 *   }
 * });
 * ```
 *
 * Note that for Angular projects, there is a harness available to help with tests that use this function in `@s-libs/ng-dev`.
 */
function isPageVisible$() {
    return fromEvent(document, 'visibilitychange').pipe(startWith(undefined), map(() => document.visibilityState === 'visible'), distinctUntilChanged());
}

/**
 * Subscribe to the returned observable to acquire a [wake lock]{@link https://developer.mozilla.org/en-US/docs/Web/API/Screen_Wake_Lock_API}. The wake lock will be re-acquired whenever the page is visible automatically. Unsubscribe to release it.
 *
 * There are no guarantees about what or when the returned observable emits.
 */
function keepWakeLock$() {
    let sentinel;
    const nav = navigator;
    return isPageVisible$().pipe(filter(isTruthy), switchMap(async () => {
        try {
            sentinel = await nav.wakeLock?.request('screen');
        }
        catch {
            // can happen when e.g. the battery is low
        }
    }), finalize(() => {
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        sentinel?.release();
    }));
}

/**
 * Mixes in {@link SubscriptionManager} as an additional superclass.
 *
 * ```ts
 * class MySubclass extends mixInSubscriptionManager(MyOtherSuperclass) {
 *   subscribeAndManage(observable: Observable<any>) {
 *     this.subscribeTo(observable);
 *   }
 * }
 * ```
 */
// eslint-disable-next-line max-lines-per-function,@typescript-eslint/explicit-function-return-type,@typescript-eslint/explicit-module-boundary-types
function mixInSubscriptionManager(Base) {
    return class extends Base {
        #subscriptions = new Subscription();
        // eslint-disable-next-line @typescript-eslint/max-params
        subscribeTo(observable, next, error, complete) {
            this.#subscriptions.add(observable.subscribe({
                next: next?.bind(this),
                error: error?.bind(this),
                complete: complete?.bind(this),
            }));
        }
        manage(subscription) {
            this.#subscriptions.add(subscription);
        }
        unsubscribe() {
            this.#subscriptions.unsubscribe();
            this.#subscriptions = new Subscription();
        }
    };
}
/**
 * Tracks all subscriptions to easily unsubscribe from them all during cleanup. Also binds callbacks to `this` for convenient use as a superclass, e.g.:
 *
 * ```ts
 * class EventLogger extends SubscriptionManager {
 *   constructor(private prefix: string, event$: Observable<string>) {
 *     super();
 *
 *     // you can pass in an instance method here and it will be bound to `this`
 *     this.subscribeTo(event$, this.log);
 *   }
 *
 *   log(event: string) {
 *     // even though this is used as a callback, you can still use `this`
 *     console.log(this.prefix + event);
 *   }
 * }
 * ```
 */
class SubscriptionManager extends mixInSubscriptionManager(Object) {
}

/*
 * Public API Surface of rxjs-core
 */

/**
 * Generated bundle index. Do not edit.
 */

export { SubscriptionManager, cache, createOperatorFunction, debounceMap, delayOnMicrotaskQueue, distinctUntilKeysChanged, filterBehavior, isPageVisible$, keepWakeLock$, logToReduxDevtoolsExtension, logValues, mapAndCacheArrayElements, mapAndCacheObjectElements, mapToLatestFrom, mixInSubscriptionManager, skipAfter, withHistory };
//# sourceMappingURL=rxjs-core.mjs.map
